/**
*	lcd.c
*	kbchCoiQO~SsΉC^[tF[X͂Srbgf[^j
*	PORT C		b0- D4
*			b1- D5
*			b2- D6
*			b3- D7
*			b4- R/~W
*			b5- RS
*	PORT D		b4- E
*	Copyright (c)Junnichi Tomaru 2007
*/
#include	<avr/io.h>
#include	"lcd.h"

extern	void	waitms( word time );

#define	WAIT_BUSY()	while(lcdbusy());

#define	LCD_ODATA	PORTC
#define	LCD_IDATA	PINC

#define ASSERT_RS( )	{ PORTC |= 0x20; }
#define ASSERT_RW( )	{ PORTC |= 0x10; }
#define ASSERT_E( )	{ PORTD |= 0x08; }

#define NEGATE_RS( )	{ PORTC &= 0xdf; }
#define NEGATE_RW( )	{ PORTC &= 0xef; }
#define NEGATE_E( )	{ PORTD &= 0xf7; }

#define WRITE_MODE( )	{ DDRC = 0x3f; }
#define READ_MODE( )	{ DDRC = 0x30; }


/**
*	f[^Zbg
*	 in	f[^
*	out	Ȃ
*/
void	write_data( byte data )
{
	byte	dat;

	dat = LCD_ODATA & 0xf0;
	dat |= data & 0x0f;
	LCD_ODATA = dat;
}


/**
*	kbcɃf[^
*	 in	f[^
*	out	Ȃ
*/
void	lcdwrite( unsigned char data )
{
	WRITE_MODE( );

	NEGATE_RW( );
	write_data( data >> 4 );
	ASSERT_E( );
	asm( "nop" );
	asm( "nop" );
	asm( "nop" );
	NEGATE_E( );
	write_data( data );
	ASSERT_E( );
	asm( "nop" );
	asm( "nop" );
	asm( "nop" );
	NEGATE_E( );
}


/**
*	kbcf[^ǂ
*	 in	Ȃ
*	out	f[^
*/
unsigned char	lcdread( void )
{
	byte	data;

	READ_MODE( );

	ASSERT_RW( );
	asm( "nop" );
	ASSERT_E( );
	asm( "nop" );
	asm( "nop" );
	data = LCD_IDATA << 4;
	NEGATE_E( );
	asm( "nop" );
	asm( "nop" );
	ASSERT_E( );
	asm( "nop" );
	asm( "nop" );
	data |= LCD_IDATA & 0x0f;
	NEGATE_E( );

	return data;
}


/**
*	CXgNV
*	 in	f[^
*	out	Ȃ
*/
void	lcdwriteir( unsigned char data )
{
	NEGATE_RS( );
	lcdwrite( data );
}


/**
*	CXgNVǂ
*	 in	Ȃ
*	out	f[^
*/
static unsigned char	lcdreadir( void )
{
	NEGATE_RS( );
	return	lcdread( );
}


/**
*	f[^WX^
*	 in	f[^
*	out	Ȃ
*/
static void	lcdwritedr( unsigned char data )
{
	ASSERT_RS( );
	lcdwrite( data );
}


/**
*	f[^WX^ǂ
*	 in	Ȃ
*	out	f[^
*/
/**//*unsigned char	lcdreaddr( void )*/
/**//*{*/
/**//*	ASSERT_RS( );*/
/**//*	return	lcdread( );*/
/**//*}*/


/**
*	CXgNViSrbgj
*	 in	f[^
*	out	Ȃ
*/
void	lcdwriteir4( unsigned char data )
{
	WRITE_MODE( );

	NEGATE_RS( );
	NEGATE_RW( );
	write_data( data );
	asm( "nop" );
	ASSERT_E( );
	asm( "nop" );
	asm( "nop" );
	asm( "nop" );
	NEGATE_E( );
}


/**
*	rW[Ԏ
*	 in	Ȃ
*	out	0x80:rW[ 0:rW[ł͂Ȃ
*/
char	lcdbusy( void )
{
	return	lcdreadir( ) & 0x80;
}


/**
*	kbc
*	 in	Ȃ
*	out	Ȃ
*/
void	lcdinit( void )
{
	waitms( 40 );
	lcdwriteir4( 0x03 );
	waitms( 5 );
	lcdwriteir4( 0x03 );
	waitms( 1 );
	lcdwriteir4( 0x03 );
	waitms( 1 );
	lcdwriteir4( 0x02 );	/* Data length 4bits */
	waitms( 1 );
	WAIT_BUSY( );
	lcdwriteir( 0x28 );	/* Set function */
	WAIT_BUSY( );
	lcdwriteir( 0x08 );	/* Display off, cursor off, blink off */
	WAIT_BUSY( );
	lcdwriteir( 0x01 );	/* Clear display */
	WAIT_BUSY( );
	lcdwriteir( 0x06 );	/* Entry mode : cursor increment */
	WAIT_BUSY( );
	lcdwriteir( 0x0C );	/* Display on, cursor off, blink off */
}


/**
*	J[\ʒuw
*	 in	x:ʒu
*		y:sʒu
*	out	Ȃ
*/
void	lcdlocate( char x, char y )
{
	char	addr;

	if( y == 0 ){
		addr = 0x00 + x;
	}
	else if( y == 1 ){
		addr = 0x40 + x;
	}
	else if( y == 2 ){
		addr = 0x14 + x;
	}
	else{
		addr = 0x54 + x;
	}
	WAIT_BUSY( );
	lcdwriteir( 0x80 | addr );	/* Set DD RAM address */
}


/**
*	ʃNA
*	 in	Ȃ
*	out	Ȃ
*/
void	lcdcls( void )
{
	WAIT_BUSY( );
	lcdwriteir( 0x01 );
}


/**
*	J[\\E\
*	 in	0:\ ̑:\
*	out	Ȃ
*/
void	lcddispcursor( char on )
{
	WAIT_BUSY( );
	lcdwriteir( 0xc | ( on ? 0x02 : 0 ) );
}


/**
*	ꕶ\
*	 in	f[^
*	out	Ȃ
*/
void	lcdputch( char c )
{
	WAIT_BUSY( );
	lcdwritedr( c );
}


/**
*	wʒuɈꕶ\iLN^R[hwj
*	 in	   x:ʒu
*		   y:sʒu
*		code:f[^
*	out	Ȃ
*/
void	lcdputcode( char x, char y, char code )
{
	lcdlocate( x, y );
	WAIT_BUSY( );
	lcdwritedr( code );
}


/**
*	\
*	 in	f[^
*	out	Ȃ
*/
void	lcdputstr( char *str )
{
	while( *str != '\0' ){
		lcdputch( *( str++ ) );
	}
}


/**
*	ʒuw肵ĕ\
*	 in	  x:ʒu
*		  y:sʒu
*		str:f[^
*	out	Ȃ
*/
void	lcdprint( char x, char y, char *str )
{
	lcdlocate( x, y );
	lcdputstr( str );
}


/**
*	ʒuw肵ČŒ蕶\
*	 in	  x:ʒu
*		  y:sʒu
*		num:f[^ʒu
*	out	Ȃ
*/
void	lcdprintc( char x, char y, const prog_char *pgm_s )
{
	char	c;

	lcdlocate( x, y );
	while(( c = pgm_read_byte( pgm_s )) != '\0' ){
		lcdputch( c );
		pgm_s++;
	}
}


/**
*	wʒuw萔̕
*	 in	  x:ʒu
*		  y:sʒu
*		cnt::
*/
/*void	lcdspace( char x, char y, char cnt )*/
/*{*/
/*	char	i;*/

/*	lcdlocate( x, y );*/
/*	for( i = 0; i < cnt; i++ ){*/
/*		lcdputch( ' ' );*/
/*	}*/
/*}*/


/**
*	sNA
*	 in	NAs
*	out	Ȃ
*/
/*void	lcdlcls( char line )*/
/*{*/
/*	lcdspace( 0, line, 16 );*/
/*}*/


/**
*	J[\P߂
*	 in	Ȃ
*	out	Ȃ
*/
void	lcdcurleft( void )
{
	WAIT_BUSY( );
	lcdwriteir( 0x10 );	/* J[\ */
}


/* end of lcd.c */
